"common library functions to acces dic catalogs"
# TODO :
# - export to txt
# - export to html

import os
import os.path
import commands
import sys
import ConfigParser
import cPickle
import zlib
import string
import re
import time
import shutil

# init variables
VERSION = '0.7'

# exceptions
DicMountException = 'DicMountException'
DicUmountException = 'DicUmountException'
DicCorruptConfigException = 'DicCorruptConfigException'
DicCorruptDiskException = 'DicCorruptDiskException'
DicNoDeviceInConfigException = 'DicNoDeviceInConfigException'
DicNoSuchDiskException = 'DicNoSuchDiskException'
DicNoSuchDeviceException = 'DicNoSuchDeviceException'
DicFileNotFoundException = 'DicFileNotFoundException'

# file-types
TYPEFILE = 0
TYPEDIR = 1

# class to handle devices
class Device:
    "class of devices"
    def __init__( self, name ):
        self.name = name
        self.path = ""
        self.runMount = 1
        self.mountProg = "/bin/mount"
        self.umountProg = "/bin/umount"

    def mount(self):
        if self.runMount:
            output = commands.getstatusoutput(self.mountProg + ' ' + self.path)
            if output[0] != 0:
                raise DicMountException, output[1]

    def umount(self):
        if self.runMount:
            output = commands.getstatusoutput(self.umountProg + ' ' + self.path)
            if output[0] != 0:
                raise DicUmountException, output[1]

    def getDeviceFile(self):
        """ Use the output of 'mount' to get the path to the physical device. """
        output = commands.getoutput(self.mountProg)
        for line in output.split('\n'):
            device, txt1, mount_point, txt2, filesystem, options = line.split()
            if mount_point == self.path:
                return device
        return None

# class of files
class File:
    "class of files"
    # defaults
    type = TYPEFILE
    filtered = ''

    def __init__( self ) :
        self.name = ''
        self.size = 0
        self.time = 0
        self.type = TYPEFILE
        self.filtered = ''

    # extract file contents
    def filter( self, path ):
        for filter in Disk.filters:
            if filter[ 'filenameregexp' ].match( path ) != None:
                try:
                    command = string.replace( filter[ 'command' ], '%f', path)
                    pipe = os.popen( command )
                    self.filtered = pipe.read()
                    # do not store errors
                    if pipe.close() != None :
                        self.filtered = ''
                except: self.filtered = ''
                break

    # string to be searched (searchmode)
    def searched( self,searchmode = 0 ):
        if searchmode == '0':
            return self.name
        elif searchmode == '1':
            return self.filtered
        else:
            return self.name + ' ' + self.filtered

class Disk:
    "this class is used for store the contents of a disk."
    filters = []

    # init members
    def __init__( self, name ):
        self.name = name
        self.comment = ''
        self.status = ''
        self.category = 'none'
        self.size = 0
        self.time = 0
        self.files = []
        self.version = VERSION
        self.device = ''
        self.index = ''

    # recursive reads a directory, adds all files to self.files
    def readDirs( self, dir ):
        self.size = self.addFiles( dir, dir )
        self.time = time.time()

    def addFiles( self, dir, rootDir ):
        size = 0L
        for file in os.listdir( dir ):
            fileWithPath = os.path.join( dir, file )
            try:
                newFile = File()
                newFile.name = fileWithPath[ len( rootDir ): ]
                newFile.size = os.path.getsize( fileWithPath )
                newFile.time = os.path.getatime( fileWithPath )
                self.index = self.index + '/' + file
                newFile.filter( fileWithPath )
                self.files.append( newFile )

                size = size + newFile.size
                if os.path.isdir( fileWithPath ) and not os.path.islink( fileWithPath ):
                    newFile.type = TYPEDIR
                    size = size + self.addFiles( fileWithPath, rootDir )
            except OSError:
                pass
        return size

    def loadInfo( self, catalogPath ) :
        baseName = os.path.join(catalogPath, self.name)
        if os.path.isfile( baseName + '.info' ):
            f = open( baseName + '.info', 'rb' )
            lines = f.readlines()
            try :
                #self.name = lines[ 0 ][ : -1 ]
                self.category = lines [ 1 ][ : -1 ]
                self.comment = lines[ 2 ][ : -1 ]
                self.status = lines[ 3 ][ : -1 ]
                self.size = string.atol( lines[ 4 ][ : -2 ] )
                self.time = string.atof( lines[ 5 ][ : -1 ] )
            except ( ValueError, IndexError ):
                raise DicCorruptDiskException
            try :
                self.version = lines[ 6 ][ : -1 ]
            except :
                # disk is version 0.4
                self.version = '0.4'
            try :
                self.device = lines[ 7 ][ : -1 ]
            except :
                # disk is version 0.5
                self.device = ''

            f.close()
        else:
            # error
            raise DicNoSuchDiskException, self.name

    # reads the indexfile from the database
    def loadIndex(  self, catalogPath ):
        baseName = os.path.join(catalogPath, self.name)
        if os.path.isfile( baseName + '.index' ):
            # not compressed
            f = open( baseName + '.index', 'rb' )
            #self.index = cPickle.load( f )
            self.index = f.read()
            f.close()
        elif os.path.isfile( baseName + '.index.gz' ):
            # compressed
            f = open( baseName + '.index.gz', 'rb' )
            compressedIndexes = f.read()
            self.index = zlib.decompress(compressedIndexes)
            #self.index = cPickle.loads(pickledIndexes)
            f.close()
        else:
            # disk < version 0.6
            pass

    def loadFiles( self, catalogPath ) :
        baseName = os.path.join(catalogPath, self.name)
        if os.path.isfile( baseName + '.disk' ):
            # not compressed
            f = open( baseName + '.disk', 'rb' )
            self.files = cPickle.load( f )
            f.close()
        elif os.path.isfile( baseName + '.disk.gz' ):
            # compressed
            f = open( baseName + '.disk.gz', 'rb' )
            compressedFiles = f.read()
            pickledFiles = zlib.decompress(compressedFiles)
            self.files = cPickle.loads(pickledFiles)
            f.close()
        else:
            # error
            raise DicNoSuchDiskException, self.name

    # reads disk from the database
    def load(  self, catalogPath ):
        self.loadInfo( catalogPath )
        self.loadIndex( catalogPath )
        self.loadFiles( catalogPath )

    def saveInfo( self, catalogPath ) :
        baseName = os.path.join(catalogPath, self.name)
        f = open ( baseName + '.info', 'wb' )
        f.write( self.name + '\n' )
        f.write( self.category + '\n' )
        f.write( self.comment + '\n' )
        f.write( self.status + '\n' )
        f.write( repr( self.size ) + '\n' )
        f.write( repr( self.time ) + '\n' )
        f.write( self.version + '\n' )
        f.write( self.device + '\n' )
        f.close()

    # save disk's indexes to the database
    def saveIndex( self, catalogPath, compressCatalog, compressionLevel ):
        baseName = os.path.join(catalogPath, self.name)
        # oops, it silently overwrites existing disks!
        if compressCatalog:
            f = open( baseName + '.index.gz', 'wb' )
            #pickledIndexes = cPickle.dumps( self.index )
            compressedIndexes = zlib.compress( self.index, compressionLevel )
            f.write( compressedIndexes )
            f.close()
        else:
            f = open( baseName + '.index', 'wb' )
            f.write( self.index )
            #cPickle.dump( self.index, f)
            f.close()

    # save disk's contents to the database
    def save( self, catalogPath, compressCatalog, compressionLevel ):
        baseName = os.path.join(catalogPath, self.name)
        # oops, it silently overwrites existing disks!
        if compressCatalog:
            f = open( baseName + '.disk.gz', 'wb' )
            pickledFiles = cPickle.dumps( self.files , 1 )
            compressedFiles = zlib.compress( pickledFiles, compressionLevel )
            f.write( compressedFiles )
            f.close()
        else:
            f = open( baseName + '.disk', 'wb' )
            cPickle.dump( self.files, f)
            f.close()
        self.saveInfo( catalogPath )
        if self.index == '' :
            self.createIndex()
        self.saveIndex( catalogPath, compressCatalog, compressionLevel )

    def searchIndex( self, content, caseSensitive ) :
        if caseSensitive == 1 :
            if string.find( self.index, content ) != -1 :
                return 1
        else :
            if string.find( string.lower( self.index ), content ) != -1 :
                return 1
        return 0

    # create indexes for old disks ( < v0.6 )
    def createIndex( self ) :
        lastSlash = 0
        for file in self.files :
            lastSlash = string.rfind( file.name, '/' )
            self.index = self.index + '/' + file.name[ lastSlash : ]

    def getFilterInfos( self, fileName ) :
        for file in self.files :
            if file.name == fileName :
                return file.filtered
        return 'Nothing found.'
class Catalog:
    "this class is used as a placeholder of all common functions. External programs should only use only methods defined here."
    def __init__( self ):
        # default values
        self.compressCatalog = 1
        self.compressionLevel = 6
        self.devices = {}
        self.defaultDevice = 'cdrom'
        self.defaultStatus = ''
        self.defaultCategory = 'none'
        self.caseSensitiveSearch = 0
        self.regexpSearch = 0
        self.longList = 0
        self.getFoundFiles = 0
        self.vfsByStatus = 0
        self.vfsByCategory = 0
        self.searchMode = 0
        self.usemagicnumber = 0

        if sys.platform[:5] == 'linux' or sys.platform[:6] == 'cygwin': 
            self.catalogPath = os.environ['HOME'] + '/.dic/'
            configFiles = [ '/etc/dic/dic.conf', os.environ[ 'HOME' ] + '/.dic/dic.conf' ]
        elif sys.platform[:3] == 'win': 
            self.catalogPath = os.path.dirname(sys.argv[0])
            configFiles = [ os.path.dirname(sys.argv[0]) + 'dic.conf' ]
        elif sys.platform[:3] == 'mac': pass
        else:
            self.exit( 'Sorry, dic cannot run on ' + os.name )
        # parses the system-wide and user's config-files
        configParser = ConfigParser.ConfigParser()
        configParser.read(configFiles)
        # try to read default values
        try:
            self.compressCatalog = configParser.getboolean( 'DEFAULT', 'compress' )
        except:
            pass
        try:
            self.compressionLevel = configParser.getint( 'DEFAULT', 'compressionLevel' )
        except:
            pass
        try:
            self.catalogPath = configParser.get( 'DEFAULT', 'catalogPath' )
            self.catalogPath = os.path.expandvars( os.path.expanduser( self.catalogPath[ : ] ) )
        except:
            pass
        try:
            self.defaultDevice = configParser.get( 'DEFAULT', 'defaultDevice' )
        except:
            pass
        try:
            self.defaultCategory = configParser.get( 'DEFAULT', 'defaultCategory' )
        except:
            pass
        try:
            self.defaultStatus = configParser.get( 'DEFAULT', 'defaultStatus' )
        except:
            pass
        try:
            self.extfsByStatus = configParser.get( 'DEFAULT', 'extfsGroupByStatus' )
        except:
            pass
        try:
            self.extfsByCategory = configParser.get( 'DEFAULT', 'extfsGroupByCategory' )
        except:
            pass

        #current values
        self.currentDevice = ''
        self.currentStatus = ''
        self.currentCategory = ''
        self.deviceIsSpecified = 0
        self.categoryIsSpecified = 0
        self.statusIsSpecified = 0

        # read devices
        if configParser.sections() != []:
            for device in configParser.sections():
                self.devices[device] = Device( device )
                # the path option is a must.
                try:
                    self.devices[device].path = configParser.get( device, 'path' )
                except ( ConfigParser.NoOptionError ):
                    raise DicCorruptConfigException, device
                # when other parameters don't exist, use defaults.
                try: self.devices[device].mountProg = configParser.get(device, 'mountProgram')
                except (ConfigParser.NoOptionError):
                    pass
                try: self.devices[device].umountProg = configParser.get( device, 'umountProgram' )
                except (ConfigParser.NoOptionError):
                    pass
                try: self.devices[device].runMount = string.atol( configParser.get( device, 'runMount' ) )
                except (ConfigParser.NoOptionError):
                    pass
        else:
            raise DicNoDeviceInConfigException

        # create catalogpath, if not exists
        if not os.path.isdir( self.catalogPath ) :
            try :
                os.makedirs( self.catalogPath )
                self.printMessage( 'Catalog-files directory created : ' + self.catalogPath )
            except :
                self.exit( 'Cannot create directory for catalog-files : ' + self.catalogPath )
        self.getfilters()

    def getfilters( self ):
        if sys.platform[:5] == 'linux' or sys.platform[:6] == 'cygwin':
            configFiles = [ '/etc/dic/dicfilter.conf', os.environ[ 'HOME' ] + '/.dic/dicfilter.conf' ]
        elif sys.platform[:3] == 'win':
            configFiles = [os.path.join(os.path.dirname(sys.argv[0]),'dicfilter.conf')]
        elif sys.platform[:3] == 'mac': pass

        configParser = ConfigParser.ConfigParser()
        configParser.read(configFiles)

        try:
            self.usemagicnumber = configParser.getint( 'DEFAULT', 'magicnumber' )
        except:
            pass

        if configParser.sections() != []:
            for section in configParser.sections():
                try:
                    filter = {}
                    filter[ 'name' ] = section
                    filter[ 'filename' ] = configParser.get( section, 'filename' )
                    filter[ 'filenameregexp' ] = re.compile( filter[ 'filename' ], re.IGNORECASE )
                    filter[ 'command' ] = configParser.get( section, 'command' )
                    # add filter to the Disk class's class attribute
                    Disk.filters.append( filter )
                except:
                    pass

    def saveDisk( self, disk ):
        self.removeDisk( disk.name )
        if self.statusIsSpecified == 1 :
            disk.status = self.currentStatus
        if self.categoryIsSpecified == 1 :
            disk.category = self.currentCategory
        disk.save( self.catalogPath, self.compressCatalog, self.compressionLevel )

    def loadDisk( self, diskName ):
        disk = Disk( diskName )
        disk.load( self.catalogPath )
        if disk.device == '' :
            disk.device = self.defaultDevice
        return disk

    def updateDisk( self, diskName, newDiskName ) :
        disk = self.loadDisk (diskName )
        if newDiskName != '' :
            self.removeDisk( disk.name )
            disk.name = newDiskName
        self.saveDisk( disk )

    # removes a disk from the database
    def removeDisk( self, name ):
        try: os.unlink( os.path.join(self.catalogPath, name) + '.disk' )
        except: pass
        try: os.unlink( os.path.join(self.catalogPath, name) + '.disk.gz' )
        except: pass
        try: os.unlink( os.path.join(self.catalogPath, name) + '.info' )
        except: pass
        try: os.unlink( os.path.join(self.catalogPath, name) + '.index' )
        except: pass
        try: os.unlink( os.path.join(self.catalogPath, name) + '.index.gz' )
        except: pass

    # creates a 'fancy' printing of a number (123304 -> 123,304)
    def fancyNumbers( self, sizeInNumber ):
        size = "%d" % sizeInNumber
        if len(size) > 6:
            return size[:-6] + "," + size[-6:-3] + "," + size[-3:]
        elif len(size) > 3:
            return size[-6:-3] + "," + size[-3:]
        else:
            return size

    # returns all available disknames (according to the given status and category)
    def diskNames( self ):
        names = []
        name = ''
        if self.categoryIsSpecified == 1 or self.statusIsSpecified == 1 :
            loadInfos = 1
        else :
            loadInfos = 0
        try:
            for file in os.listdir( self.catalogPath ):
                if file[ -5: ] == '.info' :
                    continue
                if file[ -8: ] == '.disk.gz':
                    name = file[ : -8 ]
                elif file[ -5: ] == '.disk':
                    name = file[ : -5 ]
                else :
                    continue
                if name != '' :
                    if loadInfos == 1 :
                        disk = Disk( name )
                        disk.loadInfo( self.catalogPath )
                        if ( self.categoryIsSpecified == 0 or self.currentCategory == disk.category ) and \
                            ( self.statusIsSpecified == 0 or self.currentStatus == disk.status ) :
                            names.append( name )
                    else :
                        names.append( name )
        except (OSError):
            pass
        names.sort()
        return names

    # returns all available disknames (catalog and status are ignored)
    def allDiskNames( self ):
        names = []
        name = ''
        try:
            for file in os.listdir( self.catalogPath ):
                if file[ -5: ] == '.info' :
                    continue
                if file[ -8: ] == '.disk.gz':
                    name = file[ : -8 ]
                elif file[ -5: ] == '.disk':
                    name = file[ :-5 ]
                else :
                    continue
                if name != '' :
                    names.append( name )
        except (OSError):
            pass
        names.sort()
        return names

    # search the given strings in the given disk, return the list of files
    def search( self, diskNames, content ):
        searchInIndexes = 0
        foundFiles = []
        if self.regexpSearch == 1 :
            if self.caseSensitiveSearch == 1 :
                compiledRegexp = re.compile( content )
            else :
                compiledRegexp = re.compile( content, re.IGNORECASE )
        else :
            if not self.caseSensitiveSearch :
                content = string.lower( content )
            if string.find( content, '/' ) == -1 and self.searchMode == 0 :
                searchInIndexes = 1
        for diskName in diskNames :
            disk = Disk( diskName )
            if searchInIndexes :
                disk.loadIndex( self.catalogPath )
                if disk.index != '' :
                    if disk.searchIndex( content, self.caseSensitiveSearch ) == 0 :
                        continue
            disk.loadFiles( self.catalogPath )
            if self.regexpSearch == 1 :
                for file in disk.files:
                    if compiledRegexp.match( file.searched( self.searchMode ) ) != None :
                        file.diskName = diskName
                        foundFiles.append( file )
            else :
                if self.caseSensitiveSearch == 1 :
                    for file in disk.files:
                        if string.find( file.searched( self.searchMode ), content ) != -1:
                            file.diskName = diskName
                            foundFiles.append( file )
                else:
                    for file in disk.files:
                        if string.find( string.lower( file.searched( self.searchMode ) ), content ) != -1:
                            file.diskName = diskName
                            foundFiles.append( file )
        return foundFiles

    # split the diskname and relative path of filename from the stored name
    def splitDiskNameAndPath( self, fileName ) :
        # split the file and the diskname from the path
        firstSlash = string.find( fileName, '/' )
        # if only the diskname ( with an optional '/' ) is given
        if firstSlash == -1 or firstSlash == len( fileName ) - 1 :
            diskName = fileName
            path = '/*'
        else :
            diskName = fileName[ : firstSlash ]
            path = fileName[ firstSlash : ]
        return diskName, path

    # tries to locate the given file in the cds currently in devices
    # returns the found device
    def locateFile( self, diskName, path ) :
        if self.deviceIsSpecified :
            # a device is specified, try only that
            device = self.devices[ self.currentDevice ]
        else :
            # try the device stored in the disk while added
            disk = Disk( diskName )
            disk.loadInfo( self.catalogPath )
            if disk.device == '' :
                device = self.devices[ self.defaultDevice ]
            else :
                device = self.devices[ disk.device ]
        path = device.path + path
        return device

    def getFiles( self, files, destination ) :
        files.sort()
        diskName, path = self.splitDiskNameAndPath( files[ 0 ] )
        device = self.locateFile( diskName, path )
        try :
            device.mount()
        except DicMountException :
            pass
        for fileName in files :
            diskName, path = self.splitDiskNameAndPath( fileName )
            path = device.path + path

            #if path == catalog.devices[ device ].path :
            #    path = path + '/'

            while not os.path.isfile( path ) and not os.path.isdir( path ) :
                try:
                    device.umount()
                except DicUmountException:
                    pass
                self.printMessageAndWait( 'Please insert the disk labeled \'' + diskName + '\' into the device \'' + device.name + '\' and press Enter.' )
                try:
                    device.mount()
                except DicMountException, details:
                    self.printMessage( details )

            # Check if the file already exists
            basename = os.path.basename(path)
            if not basename: # Path probably has a trailing slash on the end
                basename = os.path.basename(path[:-1])
            if os.path.isfile(os.path.join(destination, basename)) or os.path.isdir(os.path.join(destination, basename)):
                self.exit( "'%s' already exists!" % (os.path.join(destination, basename)) )

            self.printMessage( 'copying ' + path + ' to the current directory.' )
            try:
                if os.path.isfile(path):
                    shutil.copy2(path, destination)
                elif os.path.isdir(path):
                    shutil.copytree(path, os.path.join(destination, basename))
            except KeyboardInterrupt:
                self.exit( 'exiting.' )
        try:
            device.umount()
        except DicUmountException, details :
            self.exit( details )

    def viewFilterInfos( self, fileName ) :
        diskName, path = self.splitDiskNameAndPath( fileName )
        disk = self.loadDisk( diskName )
        return disk.getFilterInfos( path )

